from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

from sziszapangma.core.alignment.alignment_step import AlignmentStep
from sziszapangma.core.alignment.word import Word


class AlignmentUtil:
    @staticmethod
    def _optional_str_to_str(word: Optional[Word]) -> str:
        return word.value if word is not None else ""

    @staticmethod
    def _wer_step_to_pandas_row_lit(step: AlignmentStep) -> Tuple[str, str, str, float]:
        return (
            step.step_type.get_short_name(),
            AlignmentUtil._optional_str_to_str(step.step_words.reference_word),
            AlignmentUtil._optional_str_to_str(step.step_words.hypothesis_word),
            round(step.step_cost, 3),
        )

    @staticmethod
    def steps_to_dataframe(steps: List[AlignmentStep]) -> pd.DataFrame:
        arr = np.array([AlignmentUtil._wer_step_to_pandas_row_lit(step) for step in steps])
        return pd.DataFrame(arr, columns=["step_type", "reference", "hypothesis", "cost"])

    @staticmethod
    def get_reference_indexes_per_steps(steps: List[AlignmentStep]) -> List[int]:
        counter = 0
        indexes = []
        for step in steps:
            indexes.append(counter)
            if step.step_type.contain_reference_word():
                counter = counter + 1
        return indexes

    @staticmethod
    def get_reference_length(steps: List[AlignmentStep]) -> int:
        return sum([1 if step.step_type.contain_reference_word() else 0 for step in steps])

    @staticmethod
    def apply_weights_to_alignment(
        steps: List[AlignmentStep], weights: List[float]
    ) -> List[AlignmentStep]:
        if AlignmentUtil.get_reference_length(steps) != len(weights):
            raise Exception(
                f"Incorrect length of weights, current={len(weights)}, "
                f"required={AlignmentUtil.get_reference_length(steps)}"
            )
        reference_indexes_per_steps = AlignmentUtil.get_reference_indexes_per_steps(steps)
        return [
            steps[index].with_weight_multiplication(weights[reference_indexes_per_steps[index]])
            for index in range(len(steps))
        ]
